/*
 * A GTK+/GNOME front end for Linuxconf.
 *
 * Copyright (C) 2000, 2001 Conectiva, Inc.
 * Copyright (C) 2000 Red Hat, Inc.
 *
 * Authors: Gustavo Niemeyer <niemeyer@conectiva.com>
 *          Nalin Dahyabhai <nalin@redhat.com>
 *
 */

#include <gnome-linuxconf.h>

#define GNOME_LINUXCONF_COMMANDS "gnome-linuxconf-commands"
#define GET_REAL_FORM(x) while (GTK_IS_ALIGNMENT(x) || GTK_IS_EVENT_BOX(x)) x = GTK_OBJECT(GTK_BIN(x)->child)

static GHashTable *brush_registry = NULL;
static GHashTable *font_registry = NULL;
static GHashTable *pen_registry = NULL;
static GHashTable *dc_registry = NULL;

struct _brush_type {
	GdkColor color;
	GdkFill fill;
};

typedef struct _brush_type brush_type;

struct _pen_type {
	GdkColor color;
	int thick;
	GdkLineStyle style;
};

typedef struct _pen_type pen_type;

struct _dc_type {
	brush_type *brush;
	pen_type *pen;
};
	
typedef struct _dc_type dc_type;

struct _command {
	int type;
	int a1, a2, b1, b2, c1, c2;
	void *data;
	dc_type *dc;
	struct _command *next;
};

typedef struct _command command;

enum command_type {
	CMD_CLEAR,
	CMD_DRAWLINE,
	CMD_DRAWRECT,
	CMD_DRAWARC,
	CMD_DRAWTEXT,
	CMD_FILLRECT,
	CMD_FILLARC
};

static dc_type *get_dc(const char *id)
{
	if (dc_registry != NULL)
		return (dc_type*) g_hash_table_lookup(dc_registry, id);
	return NULL;
}

static GList *commands_get(GtkObject *object)
{
	GET_REAL_FORM(object);
	g_return_val_if_fail(GTK_IS_OBJECT(object), NULL);
	return gtk_object_get_data(object, GNOME_LINUXCONF_COMMANDS);
}

static void commands_draw(GtkObject *object)
{
	GtkWidget *widget = (GtkWidget*) object;
	GList *cmd_list;
	GdkWindow *window;
	GdkGC *gc, *gclocal;
	GdkRectangle clip;
	GdkColormap *cm;
	if (!GTK_IS_WIDGET(widget) || !GTK_WIDGET_DRAWABLE(widget))
		return;
	cmd_list = commands_get(object);
	if (cmd_list == NULL) return;
	window = widget->window;
	//clip.x = widget->allocation.x+2;
	//clip.y = widget->allocation.y+2;
	//clip.width = widget->allocation.width-4;
	//clip.height = widget->allocation.height-4;
	gc = gdk_gc_new(window);
	//gdk_gc_set_clip_rectangle(gc, &clip);
	gclocal = gdk_gc_new(window);
	gdk_gc_copy(gclocal, gc);
	cm = gdk_window_get_colormap(window);
	for (; cmd_list != NULL; cmd_list = g_list_next(cmd_list)) {
		command *cmd = (command *) cmd_list->data;
		if (cmd->dc != NULL) {
			if (cmd->type == CMD_FILLRECT || cmd->type == CMD_FILLARC) {
				if (cmd->dc->brush != NULL) {
					if (!gdk_colormap_alloc_color(cm, &cmd->dc->brush->color, FALSE, FALSE))
						g_warning("can't allocate color");
					gdk_gc_set_foreground(gclocal, &cmd->dc->brush->color);
					gdk_gc_set_background(gclocal, &cmd->dc->brush->color);
					gdk_gc_set_fill(gclocal, cmd->dc->brush->fill);
				}
			} else {
				if (cmd->dc->pen != NULL) {
					if (!gdk_colormap_alloc_color(cm, &cmd->dc->pen->color, FALSE, FALSE))
						g_warning("can't allocate color");
					gdk_gc_set_foreground(gclocal, &cmd->dc->pen->color);
					gdk_gc_set_background(gclocal, &cmd->dc->pen->color);
					gdk_gc_set_line_attributes(gclocal, cmd->dc->pen->thick, cmd->dc->pen->style, -1, -1);
				}
			}
		}
		switch (cmd->type) {
			case CMD_DRAWLINE:
				//gdk_draw_line(window,gclocal,cmd->a1+clip.x,cmd->a2+clip.y,cmd->b1+clip.x,cmd->b2+clip.y);
				gdk_draw_line(window,gclocal,cmd->a1,cmd->a2,cmd->b1,cmd->b2);
				break;
			case CMD_DRAWRECT:
				//gdk_draw_rectangle(window,gclocal,0,cmd->a1+clip.x,cmd->a2+clip.y,cmd->b1,cmd->b2);
				gdk_draw_rectangle(window,gclocal,0,cmd->a1,cmd->a2,cmd->b1,cmd->b2);
				break;
			case CMD_DRAWARC:
				//gdk_draw_arc(window,gclocal,0,cmd->a1+clip.x,cmd->a2+clip.y,cmd->b1,cmd->b2,cmd->c1,cmd->c2);
				gdk_draw_arc(window,gclocal,0,cmd->a1,cmd->a2,cmd->b1,cmd->b2,cmd->c1,cmd->c2);
				break;
			case CMD_FILLRECT:
				//gdk_draw_rectangle(window,gclocal,1,cmd->a1+clip.x,cmd->a2+clip.y,cmd->b1,cmd->b2);
				gdk_draw_rectangle(window,gclocal,1,cmd->a1,cmd->a2,cmd->b1,cmd->b2);
				break;
			case CMD_FILLARC:
				//gdk_draw_arc(window,gclocal,1,cmd->a1+clip.x,cmd->a2+clip.y,cmd->b1,cmd->b2,cmd->c1,cmd->c2);
				gdk_draw_arc(window,gclocal,1,cmd->a1,cmd->a2,cmd->b1,cmd->b2,cmd->c1,cmd->c2);
				break;
			case CMD_DRAWTEXT:
				//gdk_draw_string(window,widget->style->font,gclocal,cmd->a1+clip.x,cmd->a2+clip.y,(char*)cmd->data);
				gdk_draw_string(window,widget->style->font,gclocal,cmd->a1,cmd->a2,(char*)cmd->data);
				break;
		}
		if (cmd->dc != NULL)
			gdk_gc_copy(gclocal, gc);
	}
}

static void commands_destroy(GtkObject *object, gpointer data)
{
	GList *cmd_list;
	GET_REAL_FORM(object);
	cmd_list = commands_get(object);
	if (cmd_list != NULL) {
		for (; cmd_list != NULL; cmd_list = g_list_next(cmd_list)) {
			g_free(cmd_list->data);
		}
		g_list_free(cmd_list);
	}
	gtk_object_remove_data(object,GNOME_LINUXCONF_COMMANDS);
	gtk_signal_disconnect_by_func(object,commands_draw,NULL);
}

static void commands_register(GtkObject *object)
{
	gtk_signal_connect(object, "destroy",
			   GTK_SIGNAL_FUNC(commands_destroy), NULL);
	gtk_signal_connect_after(object, "draw",
				 GTK_SIGNAL_FUNC(commands_draw), NULL);
	gtk_signal_connect(object, "expose_event",
			   GTK_SIGNAL_FUNC(commands_draw), NULL);
}

static GList *commands_push(GtkObject *object, command *new_cmd)
{
	GList *cmd_list;
	GET_REAL_FORM(object);
	cmd_list = commands_get(object);
	g_return_val_if_fail(new_cmd != NULL, NULL);
	if (cmd_list == NULL)
		commands_register(object);
	cmd_list =  g_list_append(cmd_list, new_cmd);
	gtk_object_set_data(object, GNOME_LINUXCONF_COMMANDS, cmd_list);
	return cmd_list;
}

HANDLE_DEF(Refresh)
{
	GtkObject *object = parents_peek();
	GET_REAL_FORM(object);
	g_return_val_if_fail(GTK_IS_WIDGET(object), NULL);
	commands_draw(object);
	return NULL;
}

HANDLE_DEF(Clear)
{
	const char *dc = (const char*) g_hash_table_lookup(opts, "dc");
	GtkObject *object = parents_peek();
	GList *cmd_list;
	GET_REAL_FORM(object);

	g_return_val_if_fail(GTK_IS_WIDGET(object), NULL);

	cmd_list = commands_get(object);
	if (cmd_list != NULL)
		commands_destroy(object, NULL);

	gtk_widget_queue_clear((GtkWidget*)object);

	return NULL;
}

HANDLE_DEF(Drawline)
{
	const char *a1 = (const char*) g_list_nth_data(args, 0);
	const char *a2 = (const char*) g_list_nth_data(args, 1);
	const char *b1 = (const char*) g_list_nth_data(args, 2);
	const char *b2 = (const char*) g_list_nth_data(args, 3);
	const char *dc = (const char*) g_hash_table_lookup(opts, "dc");
	GtkObject *object = parents_peek();
	command *cmd;
	GET_REAL_FORM(object);
	
	g_return_val_if_fail(GTK_IS_WIDGET(object), NULL);
	g_return_val_if_fail(a1 != NULL, NULL);
	g_return_val_if_fail(a2 != NULL, NULL);
	g_return_val_if_fail(b1 != NULL, NULL);
	g_return_val_if_fail(b2 != NULL, NULL);

	cmd = (command *) malloc(sizeof(command));
	g_return_val_if_fail(cmd != NULL, NULL);
	
	cmd->type = CMD_DRAWLINE;
	cmd->a1 = atoi(a1);
	cmd->a2 = atoi(a2);
	cmd->b1 = atoi(b1);
	cmd->b2 = atoi(b2);
	if (dc != NULL)
		cmd->dc = get_dc(dc);
	else
		cmd->dc = NULL;

	commands_push(object, cmd);

	return NULL;
}

HANDLE_DEF(Drawrect)
{
	const char *a1 = (const char*) g_list_nth_data(args, 0);
	const char *a2 = (const char*) g_list_nth_data(args, 1);
	const char *b1 = (const char*) g_list_nth_data(args, 2);
	const char *b2 = (const char*) g_list_nth_data(args, 3);
	const char *dc = (const char*) g_hash_table_lookup(opts, "dc");
	GtkObject *object = parents_peek();
	command *cmd;
	GET_REAL_FORM(object);

	g_return_val_if_fail(GTK_IS_WIDGET(object), NULL);
	g_return_val_if_fail(a1 != NULL, NULL);
	g_return_val_if_fail(a2 != NULL, NULL);
	g_return_val_if_fail(b1 != NULL, NULL);
	g_return_val_if_fail(b2 != NULL, NULL);

	cmd = (command *) malloc(sizeof(command));
	g_return_val_if_fail(cmd != NULL, NULL);
	
	cmd->type = CMD_DRAWRECT;
	cmd->a1 = atoi(a1);
	cmd->a2 = atoi(a2);
	cmd->b1 = atoi(b1);
	cmd->b2 = atoi(b2);
	if (dc != NULL)
		cmd->dc = get_dc(dc);
	else
		cmd->dc = NULL;

	commands_push(object, cmd);

	return NULL;
}

HANDLE_DEF(Fillrect)
{
	const char *a1 = (const char*) g_list_nth_data(args, 0);
	const char *a2 = (const char*) g_list_nth_data(args, 1);
	const char *b1 = (const char*) g_list_nth_data(args, 2);
	const char *b2 = (const char*) g_list_nth_data(args, 3);
	const char *dc = (const char*) g_hash_table_lookup(opts, "dc");
	GtkObject *object = parents_peek();
	command *cmd;
	GET_REAL_FORM(object);
	
	g_return_val_if_fail(GTK_IS_WIDGET(object), NULL);
	g_return_val_if_fail(a1 != NULL, NULL);
	g_return_val_if_fail(a2 != NULL, NULL);
	g_return_val_if_fail(b1 != NULL, NULL);
	g_return_val_if_fail(b2 != NULL, NULL);

	cmd = (command *) malloc(sizeof(command));
	g_return_val_if_fail(cmd != NULL, NULL);
	
	cmd->type = CMD_FILLRECT;
	cmd->a1 = atoi(a1);
	cmd->a2 = atoi(a2);
	cmd->b1 = atoi(b1)+1;
	cmd->b2 = atoi(b2)+1;
	if (dc != NULL)
		cmd->dc = get_dc(dc);
	else
		cmd->dc = NULL;

	commands_push(object, cmd);

	return NULL;
}

HANDLE_DEF(Drawarc)
{
	const char *a1 = (const char*) g_list_nth_data(args, 0);
	const char *a2 = (const char*) g_list_nth_data(args, 1);
	const char *b1 = (const char*) g_list_nth_data(args, 2);
	const char *b2 = (const char*) g_list_nth_data(args, 3);
	const char *c1 = (const char*) g_list_nth_data(args, 4);
	const char *c2 = (const char*) g_list_nth_data(args, 5);
	const char *dc = (const char*) g_hash_table_lookup(opts, "dc");
	GtkObject *object = parents_peek();
	command *cmd;
	GET_REAL_FORM(object);
	
	g_return_val_if_fail(GTK_IS_WIDGET(object), NULL);
	g_return_val_if_fail(a1 != NULL, NULL);
	g_return_val_if_fail(a2 != NULL, NULL);
	g_return_val_if_fail(b1 != NULL, NULL);
	g_return_val_if_fail(b2 != NULL, NULL);
	g_return_val_if_fail(c1 != NULL, NULL);
	g_return_val_if_fail(c2 != NULL, NULL);

	cmd = (command *) malloc(sizeof(command));
	g_return_val_if_fail(cmd != NULL, NULL);

	cmd->type = CMD_DRAWARC;
	cmd->a1 = atoi(a1);
	cmd->a2 = atoi(a2);
	cmd->b1 = atoi(b1);
	cmd->b2 = atoi(b2);
	cmd->c1 = atoi(c1);
	cmd->c2 = atoi(c2);
	if (dc != NULL)
		cmd->dc = get_dc(dc);
	else
		cmd->dc = NULL;

	commands_push(object, cmd);

	return NULL;
}

HANDLE_DEF(Fillarc)
{
	const char *a1 = (const char*) g_list_nth_data(args, 0);
	const char *a2 = (const char*) g_list_nth_data(args, 1);
	const char *b1 = (const char*) g_list_nth_data(args, 2);
	const char *b2 = (const char*) g_list_nth_data(args, 3);
	const char *c1 = (const char*) g_list_nth_data(args, 4);
	const char *c2 = (const char*) g_list_nth_data(args, 5);
	const char *dc = (const char*) g_hash_table_lookup(opts, "dc");
	GtkObject *object = parents_peek();
	command *cmd;
	GET_REAL_FORM(object);
	
	g_return_val_if_fail(GTK_IS_WIDGET(object), NULL);
	g_return_val_if_fail(a1 != NULL, NULL);
	g_return_val_if_fail(a2 != NULL, NULL);
	g_return_val_if_fail(b1 != NULL, NULL);
	g_return_val_if_fail(b2 != NULL, NULL);
	g_return_val_if_fail(c1 != NULL, NULL);
	g_return_val_if_fail(c2 != NULL, NULL);

	cmd = (command *) malloc(sizeof(command));
	g_return_val_if_fail(cmd != NULL, NULL);

	cmd->type = CMD_FILLARC;
	cmd->a1 = atoi(a1);
	cmd->a2 = atoi(a2);
	cmd->b1 = atoi(b1);
	cmd->b2 = atoi(b2);
	cmd->c1 = atoi(c1);
	cmd->c2 = atoi(c2);
	if (dc != NULL)
		cmd->dc = get_dc(dc);
	else
		cmd->dc = NULL;

	commands_push(object, cmd);

	return NULL;
}

HANDLE_DEF(Drawtext)
{
	const char *a1 = (const char*) g_list_nth_data(args, 0);
	const char *a2 = (const char*) g_list_nth_data(args, 1);
	const char *data = (const char*) g_list_nth_data(args, 2);
	const char *dc = (const char*) g_hash_table_lookup(opts, "dc");
	GtkObject *object = parents_peek();
	command *cmd;
	GET_REAL_FORM(object);
	
	g_return_val_if_fail(GTK_IS_WIDGET(object), NULL);
	g_return_val_if_fail(a1 != NULL, NULL);
	g_return_val_if_fail(a2 != NULL, NULL);
	g_return_val_if_fail(data != NULL, NULL);

	cmd = (command *) malloc(sizeof(command));
	g_return_val_if_fail(cmd != NULL, NULL);
	
	cmd->type = CMD_DRAWTEXT;
	cmd->a1 = atoi(a1);
	cmd->a2 = atoi(a2);
	cmd->data = (void*) g_strdup(data);
	if (dc != NULL)
		cmd->dc = get_dc(dc);
	else
		cmd->dc = NULL;

	commands_push(object, cmd);

	return NULL;
}

HANDLE_DEF(Defbrush)
{
	const char *id = (const char*) g_list_nth_data(args, 0);
	const char *color = (const char*) g_list_nth_data(args, 1);
	const char *style = (const char*) g_list_nth_data(args, 2);
	brush_type *brush;

	g_return_val_if_fail(id != NULL, NULL);
	g_return_val_if_fail(color != NULL, NULL);
	g_return_val_if_fail(style != NULL, NULL);
	
	brush = (brush_type*) malloc(sizeof(brush_type));
	if (!gdk_color_parse(color, &brush->color))
		g_warning("Defbrush: can't parse color '%s'", color);
	brush->fill = GDK_SOLID;

	if(brush_registry == NULL)
		brush_registry = g_hash_table_new(g_str_hash, g_str_equal);
	
	g_hash_table_insert(brush_registry, g_strdup(id), brush);
	
	return NULL;
}

HANDLE_DEF(Defpen)
{
	const char *id = (const char*) g_list_nth_data(args, 0);
	const char *color = (const char*) g_list_nth_data(args, 1);
	const char *thick = (const char*) g_list_nth_data(args, 2);
	const char *style = (const char*) g_list_nth_data(args, 3);
	int istyle;
	pen_type *pen;

	g_return_val_if_fail(id != NULL, NULL);
	g_return_val_if_fail(color != NULL, NULL);
	g_return_val_if_fail(thick != NULL, NULL);
	g_return_val_if_fail(style != NULL, NULL);
	
	pen = (pen_type*) malloc(sizeof(pen_type));
	if (!gdk_color_parse(color, &pen->color))
		g_warning("Defpen: can't parse color '%s'", color);
	pen->thick = atoi(thick);
	istyle = atoi(style);
	switch(istyle) {
		case 0:
		case 1:
			pen->style = GDK_LINE_SOLID;
			break;
		case 101:
		case 102:
		case 103:
		case 104:
			pen->style = GDK_LINE_ON_OFF_DASH;
			break;
		default:
			pen->style = GDK_LINE_SOLID;
			g_warning("Defpen: unrecognized pen style '%i'", istyle);
	}

	if(pen_registry == NULL)
		pen_registry = g_hash_table_new(g_str_hash, g_str_equal);
	
	g_hash_table_insert(pen_registry, g_strdup(id), pen);
	
	return NULL;
}

HANDLE_DEF(Defdc)
{
	const char *id = (const char*) g_list_nth_data(args, 0);
	const char *brushid = (const char*) g_hash_table_lookup(opts, "brush");
	const char *penid = (const char*) g_hash_table_lookup(opts, "pen");
	dc_type *dc;

	g_return_val_if_fail(id != NULL, NULL);
	
	dc = (dc_type*) malloc(sizeof(dc_type));
	if (brushid != NULL) {
		dc->brush = g_hash_table_lookup(brush_registry, brushid);
	} else {
		dc->brush = NULL;
	}
	if (penid != NULL) {
		dc->pen = g_hash_table_lookup(pen_registry, penid);
	} else {
		dc->pen = NULL;
	}

	if(dc_registry == NULL)
		dc_registry = g_hash_table_new(g_str_hash, g_str_equal);
	
	g_hash_table_insert(dc_registry, g_strdup(id), dc);
	
	return NULL;
}

