/*
 * A GTK+/GNOME front end for Linuxconf.
 *
 * Copyright (C) 2000, 2001 Conectiva, Inc.
 * Copyright (C) 2000 Red Hat, Inc.
 *
 * Authors: Gustavo Niemeyer <niemeyer@conectiva.com>
 *          Nalin Dahyabhai <nalin@redhat.com>
 *
 */

#include "gnome-linuxconf.h"

/* The previous widget created. */
GtkObject *previous_object = NULL;

/* Simple globally-accessible parent "stack". */
static GList *parents_stack = NULL;

GtkObject *parents_push(GtkObject *new_parent)
{
	g_return_val_if_fail(new_parent != NULL, NULL);
	g_return_val_if_fail(GTK_IS_OBJECT(new_parent), NULL);
	parents_stack = g_list_prepend(parents_stack, new_parent);
	return new_parent;
}

GtkObject *parents_pop()
{
	GtkObject *ret = GTK_OBJECT(g_list_nth_data(parents_stack, 0));
	g_return_val_if_fail(parents_stack != NULL, NULL);
	parents_stack = g_list_remove(parents_stack, ret);
	return ret;
}

GtkObject *parents_peek()
{
	return (GtkObject*) g_list_nth_data(parents_stack, 0);
}

/* Simple globally-accessible mainform "stack". */
static GList *mainforms_stack = NULL;

GtkObject *mainforms_push(GtkObject *new_mainform)
{
	g_return_val_if_fail(new_mainform != NULL, NULL);
	g_return_val_if_fail(GTK_IS_OBJECT(new_mainform), NULL);
	mainforms_stack = g_list_prepend(mainforms_stack, new_mainform);
	return new_mainform;
}

GtkObject *mainforms_pop()
{
	GtkObject *ret = GTK_OBJECT(g_list_nth_data(mainforms_stack, 0));
	g_return_val_if_fail(mainforms_stack != NULL, NULL);
	mainforms_stack = g_list_remove(mainforms_stack, ret);
	return ret;
}

GtkObject *mainforms_peek()
{
	return (GtkObject*) g_list_nth_data(mainforms_stack, 0);
}

/* A simple globally-accessible "stack" of tree items. */
static GList *tree_stack = NULL;

GtkCTreeNode *tree_push(GtkCTreeNode *new_parent)
{
	g_return_val_if_fail(new_parent != NULL, NULL);
	tree_stack = g_list_prepend(tree_stack, new_parent);
	return new_parent;
}

GtkCTreeNode *tree_pop()
{
	GtkCTreeNode *ret = (GtkCTreeNode*) g_list_nth_data(tree_stack, 0);
	g_return_val_if_fail(tree_stack != NULL, NULL);
	tree_stack = g_list_remove(tree_stack, ret);
	return ret;
}

GtkCTreeNode *tree_peek()
{
	return (GtkCTreeNode*) g_list_nth_data(tree_stack, 0);
}

/* Functions to handle object accesses by name. */
static GHashTable *object_registry = NULL;

void unregister_object(GtkObject *object, const char *cid)
{
	GtkObject *registered;
	registered = (GtkObject*)g_hash_table_lookup(object_registry, cid);
	if(registered == object) {
#ifdef DEBUG
		printf("Unregistering \"%s\".\n", cid);
#endif
		g_hash_table_remove(object_registry, cid);
	}
}

GtkObject *get_object_mainform(GtkObject *object)
{
	return gtk_object_get_data(object, GNOME_LINUXCONF_MAINFORM);
}

const char *get_object_name(GtkObject *object)
{
	return (const char*)gtk_object_get_data(object, GNOME_LINUXCONF_NAME);
}

const char *get_object_base_name(GtkObject *object)
{
	const char *name = get_object_name(object);
	const char *last_dot;
	g_return_val_if_fail(name != NULL, NULL);
	last_dot = strrchr(name, '.');
	if (last_dot != NULL) {
		name = last_dot + 1;
	}
	return name;
}

char *get_object_path(GtkObject *object)
{
	const char *name = get_object_name(object);
	char *path;
	char *last_dot;
	g_return_val_if_fail(name != NULL, NULL);
	path = g_strdup(name);
	last_dot = strrchr(path, '.');
	if (last_dot != NULL) {
		*last_dot = 0;
	} else {
		g_free(path);
		path = "";
	}
	return path;
}

const char *get_object_relative_name(GtkObject *object)
{
	GtkObject *mainform = get_object_mainform(object);
	int l;

	g_return_val_if_fail(GTK_IS_OBJECT(mainform), "");
	l = strlen(get_object_name(mainform)) -
	    strlen(get_object_base_name(mainform));

	return get_object_name(object) + l;
}

int get_object_type(GtkObject *object)
{
	gpointer type = gtk_object_get_data(object, GNOME_LINUXCONF_TYPE);
	return (type == NULL) ? -1 : GPOINTER_TO_INT(type);
}

void register_object_with_parent(GtkObject *object, const char *id)
{
       	const char *pid = NULL;

	g_return_if_fail(id != NULL);
	
	/* Get the name of the widget's parent on the parent stack.  Often,
	 * the widget has already been pushed onto the stack, so check for
	 * that case and handle it properly.  */
	if(parents_peek() == object) {
		if(g_list_next(parents_stack) != NULL) {
			pid = get_object_name(g_list_next(parents_stack)->data);
		}
	} else {
		pid = get_object_name(parents_peek());
	}

	register_object(object, pid, id);
}

void register_object(GtkObject *object, const char *path, const char *id)
{
	char *cid;
	g_return_if_fail(id != NULL);
	
	/* Now compose the name Linuxconf refers to it as. */
	if(path != NULL) {
		cid = g_strdup_printf("%s.%s", path, id);
	} else {
		cid = g_strdup(id);
	}

	/* Initialize the widget registry if we need to. */
	if(object_registry == NULL) {
		object_registry = g_hash_table_new(g_str_hash, g_str_equal);
	}

#ifdef DEBUG
	printf("Registering \"%s\".\n", cid);
#endif
	/* Insert it, and add a destroy handler to free the widget's name. */
	g_hash_table_insert(object_registry, cid, object);
	gtk_object_set_data(object, GNOME_LINUXCONF_NAME, cid);
	gtk_signal_connect(object, "destroy", GTK_SIGNAL_FUNC(unregister_object), cid);
	gtk_signal_connect(object, "destroy", GTK_SIGNAL_FUNC(free_data), cid);
}

GtkObject *get_object_by_name(const char *cid)
{
	if(object_registry == NULL) {
		return NULL;
	} else {
		return g_hash_table_lookup(object_registry, cid);
	}
}

GtkObject *get_object_by_name_with_parent(const char *parentid, const char *id)
{
	GtkObject *object;
	const char *pid;
	char *cid;
	if (parentid == NULL) {
		pid = get_object_name(parents_peek());
	} else {
		pid = parentid;
	}
	g_return_val_if_fail(pid != NULL, NULL);
	cid = g_strdup_printf("%s.%s", pid, id);
	object = get_object_by_name(cid);
	g_free(cid);
	return object;
}

/* Manipulate the list of opened windows. */
static GList *window_list = NULL;

void window_list_add(GtkWidget *window, gpointer ignored)
{
	window_list = g_list_prepend(window_list, window);
}

void window_list_remove(GtkWidget *window, gpointer ignored)
{
	window_list = g_list_remove(window_list, window);
}

static GdkCursor *cursor = NULL;

void window_list_unbusy_all(GtkWidget *unused, gpointer ignored)
{
	GList *win = window_list;
	while((win != NULL) && GTK_IS_WINDOW(win->data)) {
		if((GTK_WIDGET(win->data))->window != NULL) {
			gdk_window_set_cursor((GTK_WIDGET(win->data))->window,
				       	      NULL);
		}
		win = g_list_next(win);
	}
}

void window_list_busy_all(GtkWidget *unused, gpointer ignored)
{
	GList *win = window_list;
	while((win != NULL) && GTK_IS_WINDOW(win->data)) {
		if(cursor == NULL) {
			cursor = gdk_cursor_new(GDK_WATCH);
		}
		if(((GTK_WIDGET(win->data))->window != NULL)&&(cursor != NULL)){
			gdk_window_set_cursor((GTK_WIDGET(win->data))->window,
					      cursor);
		}
		gdk_window_set_cursor((GTK_WIDGET(win->data))->window, cursor);
		win = g_list_next(win);
	}
}

int window_list_length(void)
{
	return g_list_length(window_list);
}

/* Functions to handle XPM objects by name. */
static GHashTable *xpm_object_registry = NULL;

void unregister_xpm_object(GtkObject *object, const char *id)
{
	GtkObject *registered;
	registered = (GtkObject*)g_hash_table_lookup(xpm_object_registry, id);
	if(registered == object) {
#ifdef DEBUG
		printf("Unregistering XPM \"%s\".\n", id);
#endif
		g_hash_table_remove(xpm_object_registry, id);
	}
}

void register_xpm_object(GtkObject *object, const char *id)
{
	char *did;
	g_return_if_fail(id != NULL);

	/* Initialize the widget registry if we need to. */
	if(xpm_object_registry == NULL) {
		xpm_object_registry = g_hash_table_new(g_str_hash, g_str_equal);
	}

#ifdef DEBUG
	printf("Registering XPM \"%s\".\n", id);
#endif
	did = g_strdup(id);
	/* Insert it, and add a destroy handler to free the widget's name. */
	g_hash_table_insert(xpm_object_registry, did, object);
	gtk_object_set_data(object, GNOME_LINUXCONF_NAME, did);
	gtk_signal_connect(object, "destroy", GTK_SIGNAL_FUNC(unregister_xpm_object), did);
	gtk_signal_connect(object, "destroy", GTK_SIGNAL_FUNC(free_data), did);
}

GtkObject *get_xpm_object_by_name(const char *id)
{
	if(xpm_object_registry == NULL) {
		return NULL;
	} else {
		return g_hash_table_lookup(xpm_object_registry, id);
	}
}

